#include <qregexp.h>
#include <qptrstack.h>
#include <qstring.h>
#include <kate/document.h>

#include "cppdevelop.utils.h"
#include "class.cppview.h"
#include "class.objectmap.h"
#include "class.absparser.h"

static const char *defNameByTypeC[] = { "fn",
                                                             "class",
                                                             "macro",
                                                              "struct",
                                                               NULL };

static const char *needFieldC[] = {  "@brief\n * ",
                                                     "@author Andrei Dmitriev <andrei_dm@mail.ru>",
                                                     NULL
                                                  };

typedef QPtrStack<DoxyObj> Stack;

/*!
 * @fn AbsParser::addObject
 * @param QString &name
 * @param int strNum
 * @param CppView::DefType type
 * @return void
 * @brief add Element in CppView
 * @author Andrei Dmitriev <andrei_dm@mail.ru>
 */
void AbsParser::addObject(QString &name, int strNum, CppView::DefType type)
{
        ObjectMap *objMap = ObjectMap::Instance();
        name = name.replace("\n"," ");
        logger("name parse:" + name);
        ItemObject  *obj = objMap->addObject(name, type);
        if(obj) obj->setStrNum(strNum);
        objMap->FreeInst();
}

/*!
 * @fn AbsParser::addStruct
 * @param QString &name
 * @param int strNum
 * @return void
 * @brief API for child classes for add Struct
 * @author Andrei Dmitriev <andrei_dm@mail.ru>
 */
void AbsParser::addStruct(QString &name, int strNum)
{
        name.replace(QRegExp("(typedef)?\\s*struct\\s*"),"");
        addObject(name, strNum, CppView::Struct );
}

/*!
 * @fn AbsParser::addClass
 * @param QString &name
 * @param int strNum
 * @return void
 * @brief API for child classes for add Class
 * @author Andrei Dmitriev <andrei_dm@mail.ru>
 */
void AbsParser::addClass(QString &name, int strNum)
{
        name.replace(QRegExp("class\\s*"),"");
        addObject(name, strNum, CppView::Class );
}

/*!
 * @fn ParseUtils::shiftStr
 * @param uint strPos
 * @return void
 * @brief Description
 * shift curStr for parse
 * @author Andrei Dmitriev <andrei_dm@mail.ru>
 */
void ParseUtils::shiftStr(uint strPos)
{
           if(strPos < (curStr.length())) curStr = curStr.mid(strPos);
           else curStr = QString("");
            curStr = curStr.stripWhiteSpace();
}

/*!
 * @fn CppParser::createDoc
 * @param const QString &name
 * @param CppView::DefType type
 * @return QString
 * @overload AbsParser::createDoc
 * @brief Description
 * generate Doxy document in Qt style overload virtal fn for Class AbsParser
 * @author Andrei Dmitriev <andrei_dm@mail.ru>
 */
QString CppParser::createDoc(const QString &name, CppView::DefType type)
{
        QString def = QString(defNameByTypeC[type]);
        QString nameForDoc;
        if(type == CppView::Func)
        {
            QRegExp rxFnName("^\\s*([0-9A-Za-z\\-:_&\\*~]+)\\s*([0-9A-Za-z\\-:_~&\\*]+)?\\s*\\(\\s*([^\\)]*)");
            // 1 return 2 - name 3 - param
            if(rxFnName.search(name) != -1) {
                  // step 1 - find fn name
                  QString tmpName = rxFnName.cap(2);
                  nameForDoc = rxFnName.cap( (tmpName.isEmpty()) ? 1 : 2);
                  // step 2 - parse param
                  QString param = rxFnName.cap(3);
                  if(!param.isEmpty()) {
                     nameForDoc.append("\n * ");
                     QStringList list = QStringList::split(QRegExp("\\s*,\\s*"), param);
                     nameForDoc.append(QString("@param ") + list.join(QString("\n * @param ")));
                  }
                  // step 3 add return if exist
                  if(!tmpName.isEmpty()) {
                    QString ret = rxFnName.cap(1);
                    nameForDoc.append("\n * @return " + ret);
                  }
            }
        } else nameForDoc.append(name);

        QString retStr = QString("/*!\n * @%1 %2 \n */\n");
        // add need param
        int i = 0;
        while(needFieldC[i] != NULL) nameForDoc.append(QString("\n * ")+QString(needFieldC[i++]));
        retStr = retStr.arg(def).arg(nameForDoc);
        return retStr;
}

/*!
 * @fn CppParser::parse
 * @param Kate::Document *doc
 * @return void
 * @brief overload virtual function @see class AbsParser for detail
 * @author Andrei Dmitriev <andrei_dm@mail.ru>
 */
void CppParser::parse(Kate::Document *doc)
{
        int i, curI, posDef, inTag, posStr, iFirst, defDeclaration = 0;
        int comment = 0;
        int qoute = 0;
        int blockF = 0;
        int blockR = 0;
        int lastCommentStrNum = 0;
        inTag = 0;
        curI = 0;
        enum TagType {Empty=0, Macro, Func, Class, Struct } imIN = Empty;
        QString name, tmpStr, DoxyComment;
        Stack doxyStack;
	doxyStack.setAutoDelete(FALSE);
        QRegExp rxMacro( "^#\\s*define\\s+([^\\s\\r\\n]+)" );
        QRegExp rxClass( "^\\s*((template\\s*<class[^>]*>\\s*)?class\\s*[^\\{;]*)" );
        QRegExp rxStruct( "^\\s*((typedef\\s*)?struct\\s*[^\\{]*)" );
        QRegExp rxDoxy("@(fn|struct|class|macro)\\s+([0-9a-zA-Z_\\-:]+)");
        for (i=0; i<doc->numLines();)
        {
            if(curI == 0 || (curI != i) || curStr.isEmpty()) {
                if(i != 0 && curI == i && curStr.isEmpty()) i++;
                if(i >= doc->numLines()) return;
		curStr = doc->textLine(i);
                curStr = curStr.replace(0x9,0x20);
            }
            curStr = curStr.stripWhiteSpace();
            curI = i;
            // delete comment && quote
            if((curStr.find("\"") !=-1 || curStr.find("'")!=-1)) { // qoute
                    curStr = curStr.replace("\\\"", "");
                    curStr = curStr.replace("\\'", "");
                    curStr = curStr.replace(QRegExp("\"[^\"]*\""), "");
                    curStr = curStr.replace(QRegExp("'[^']*'"), "");
            }
            if(!comment && (curStr.find("//") != -1 || curStr.find("/*")!= -1)) { // comment
                    curStr = curStr.replace(QRegExp("/\\*.*\\*/"), "");
                    int PosOneCom = curStr.find("//");
                    posStr = curStr.find("/*");
                    if(PosOneCom != -1 && (posStr == -1 || PosOneCom < posStr )) {
                               curStr = curStr.left(PosOneCom);
                    }
                    if(posStr != -1 && ( PosOneCom == -1 || posStr < PosOneCom ) ) {
                            comment = 1;
                            DoxyComment = curStr.mid(posStr);
                            DoxyComment.append("\n");
                            curStr = curStr.left(posStr);
                    }
              } else if( comment ) {
                    if((posStr = curStr.find("*/")) != -1) {
                                DoxyComment.append(curStr.left(posStr+2));
                                shiftStr(posStr+2);
                                comment = 0;
                                if(rxDoxy.search(DoxyComment) != -1) {
                                        QString txt = rxDoxy.cap(2);
					DoxyObj *dxy = new DoxyObj(txt, DoxyComment);
                                        QString why = rxDoxy.cap(1);
                                        if(why == "fn")
                                                    dxy->type = CppView::Func;
                                        else if(why == "class")
                                                    dxy->type = CppView::Class;
                                        else if(why == "struct")
                                                     dxy->type = CppView::Struct;
                                        else if(why == "macro")
                                                     dxy->type = CppView::Macro;
                                        doxyStack.push(dxy);
                                } else {
                                          lastCommentStrNum = i;
                                }
                                continue;
                    } else {
                                DoxyComment.append(curStr);
                                DoxyComment.append("\n");
                    }
                    GOTO_NEXT;
            }
            // end delete comment
            switch(imIN)
            {
                case Macro:
                        if(!inTag) {
                                posDef = rxMacro.search( curStr );
                                if ( posDef > -1 ) {
                                        name = rxMacro.cap( 1 );
                                        shiftStr(posDef+name.length());
                                        if(lastCommentStrNum+1 ==  defDeclaration) {
                                         // doxy gen yes ( need fix )
                                            DoxyObj *dxy = new DoxyObj(name, DoxyComment);
                                            dxy->type = CppView::Macro;
                                            doxyStack.push(dxy);
                                        }
                                        addMacros(name, defDeclaration);
                                }
                        if(curStr.at(curStr.length() - 1) == '\\') inTag = 1;
                        else { imIN  =  Empty; GOTO_NEXT; }
                        }else if(inTag) {
                               if(curStr.at(curStr.length() - 1) != '\\') {
                                    imIN  =  Empty;
                                    inTag = 0;
                                    GOTO_NEXT;
                               }
                        }
                break;
                case Func:
                        if(!inTag) {
                                if(iFirst) {
                                        // find  name fn
                                        posDef =  curStr.find("(");
                                        if(posDef > -1) {
                                        tmpStr = curStr.left(posDef).stripWhiteSpace();
                                            if(!tmpStr.isEmpty()) {
                                                name =  tmpStr;
                                                if( name.find(QRegExp("^\\s*(template\\s*<class[^>]*>\\s*)?[0-9A-Za-z\\-:_]+\\s+[0-9A-Za-z\\-:_]+")) == -1 ) {
                                                            //  find Name Fn
                                                            name = findFnName(doc, name, i);
                                                }
                                             } else {
                                                name = QString("");
                                                name = findFnName(doc, name, i);
                                            }
                                            name.append(QString(" ("));
                                            shiftStr(posDef+1);
                                        }
                                        iFirst = 0;
                                }
                                posStr = curStr.find("{");
                                posDef = curStr.find(";");
                                if(posStr != -1 || (posDef != -1)) {
                                    if(posStr != -1) name.append(curStr.left(posStr));
                                    else name.append(curStr.left(posDef));
                                    if(lastCommentStrNum+1 ==  defDeclaration) {
                                         // doxy gen yes ( need fix )
                                            DoxyObj *dxy = new DoxyObj(name, DoxyComment);
                                            dxy->type = CppView::Func;
                                            doxyStack.push(dxy);
                                    }
                                    addFunc( name,  defDeclaration );
                                    if(posDef != -1 && (posStr == -1 || posStr > posDef)) { imIN = Empty; shiftStr(posDef+1); continue;  }
                                    if(posStr != -1 && (posDef == -1 || posStr < posDef)) { inTag = 1; blockR = 1; shiftStr(posStr+1); continue;  }
                                 } else {
                                    name.append(curStr);

                                 }
                        } else{
                                // goto end fn
                                GOTO_END;
                        }
                break;
                case Class:
                        if(!inTag) {
                                // class not init;
                                posDef = rxClass.search( curStr );
                                if ( posDef > -1 ) {
                                        name = rxClass.cap( 1 );
                                        shiftStr(posDef+name.length());
                                }
                                if((posStr = curStr.find("{")) != -1) {
                                        // begin implement class
                                        name.append(curStr.left(posStr));
                                        if(lastCommentStrNum+1 ==  defDeclaration) {
                                         // doxy gen yes ( need fix )
                                            DoxyObj *dxy = new DoxyObj(name, DoxyComment);
                                            dxy->type = CppView::Class;
                                            doxyStack.push(dxy);
                                        }
                                        addClass(name, defDeclaration);
                                        shiftStr(posStr+1);
                                        inTag = 1;
                                        blockR = 1;
                                        continue;
                                }
                                if((posStr = curStr.find(";")) != -1) {
                                        name.append(curStr.left(posStr));
                                        if(lastCommentStrNum+1 ==  defDeclaration) {
                                         // doxy gen yes ( need fix )
                                            DoxyObj *dxy = new DoxyObj(name, DoxyComment);
                                            dxy->type = CppView::Class;
                                            doxyStack.push(dxy);
                                        }
                                        addClass(name, defDeclaration);
                                        shiftStr(posStr+1);
                                        imIN = Empty;
                                        inTag = 0;
                                }
                        }else{
                               // I'm in class
                               GOTO_END;

                        }
                break;
                case Struct:
                       if(!inTag) {
                                posDef = rxStruct.search( curStr );
                                if ( posDef > -1 ) {
                                        name = rxStruct.cap( 1 );
                                        shiftStr(posDef+name.length());
                                }
                                if((posStr = curStr.find("{")) != -1) {
                                        // where typedef mystruct {
                                        //    void (*func)();
                                        // }struct;
                                        name.append(curStr.left(posStr));
                                        shiftStr(posStr+1);
                                        inTag = 1;
                                        blockR = 1;
                                        continue;
                                }
                                if((posStr = curStr.find(";")) != -1) {
                                        // where typedef mystruct struct;
                                        name.append(QString(" ")+curStr.left(posStr));
                                        shiftStr(posStr+1);
                                        if(lastCommentStrNum+1 ==  defDeclaration) {
                                         // doxy gen yes ( need fix )
                                            DoxyObj *dxy = new DoxyObj(name, DoxyComment);
                                            dxy->type = CppView::Struct;
                                            doxyStack.push(dxy);
                                        }
                                        addStruct(name, defDeclaration);
                                        imIN = Empty;
                                        continue;
                                }
                        }else{
                               // i'm in Macros
                               GOTO_END;

                        }
                break;
                default:
                         if(curStr.find("define ") >=0 && curStr.find(QRegExp("^#\\s*define\\s"), 0) >= 0)
                         {
                                imIN = Macro;
                                defDeclaration = i;
                                continue;
                         }else if(curStr.find("class ") > -1 && curStr.find(QRegExp("^\\s*(template\\s*<class[^>]*>\\s*)?class\\s"), 0) >= 0) {
                                imIN = Class; defDeclaration = i; continue;
                         } else if(curStr.find("struct") > -1 && curStr.find(QRegExp("^\\s*(typedef\\s*)?struct\\s"),0) > -1) {
                                imIN = Struct; defDeclaration = i; continue;
                         } else if(curStr.find("(") > -1) {
                                 iFirst = 1; imIN = Func; defDeclaration = i; continue;
                         }
                break;
            };
            i++;
        }
        ObjectMap *objMap = ObjectMap::Instance();
        DoxyObj *dxy = NULL;
	if(!doxyStack.isEmpty()){
            while((dxy = doxyStack.pop())) {
		    ItemObject  *obj = objMap->wordInObject(dxy->getName(), dxy->type);
                    if(obj) obj->setDoc(dxy); else delete dxy;
            }
	}
        objMap->FreeInst();
}

/*!
 * @fn CppParser::findFnName
 * @param Kate::Document *doc
 * @param QString name - end part function name
 * @param int pos - col in string where begin name
 * @return QString - full name function
 * @brief Description
 * find Name function if fn defined on 2 or more line doc
 * @author Andrei Dmitriev <andrei_dm@mail.ru>
 */
QString CppParser::findFnName(Kate::Document *doc, QString name, int pos)
{
    int i, posStr;
    int comment = 0;
    QString curStr1, tmpName;
    for (i=pos-1;i>=0;i--)
    {
            curStr1 = doc->textLine(i);
            curStr1 = curStr1.replace(0x9,0x20);
            curStr1 = curStr1.stripWhiteSpace();
            // delete comment
            if(!comment && (curStr1.find("//")!=-1 || curStr1.find("*/")!= -1)) {
                    return name;
                    curStr1 = curStr1.replace(QRegExp("/\\*.*\\*/"), "");
                    int PosOneCom = curStr1.find("//");
                    posStr = curStr1.find("*/");
                    if(PosOneCom != -1 && (posStr == -1 || PosOneCom < posStr )) {
                               curStr1 = curStr1.left(PosOneCom);
                               curStr1 = curStr1.stripWhiteSpace();
                    }
                    if(posStr != -1 && ( PosOneCom == -1 || posStr < PosOneCom ) ) {
                            comment = 1;
                            curStr1 = curStr1.mid(posStr+2);
                            curStr1 = curStr1.stripWhiteSpace();
                    }
              }else if( comment ) {
                    if((posStr = curStr1.find("/*")) != -1) {
                                curStr1 = curStr1.left(posStr);
                                curStr1 = curStr1.stripWhiteSpace();
                                comment = 0;
                    }
                    continue;
            } // end delete comment
            if(curStr1.isEmpty() || curStr1.find("}") != -1 || curStr1.find(";") != -1 || curStr1.find("#") != -1 || curStr1.find(">") != -1 || curStr1.find("\"") != -1) {
                    return name;
            }
            tmpName = curStr1 + QString(" ") + name;
            if(tmpName.find(QRegExp("^\\s*(template\\s*<class[^>]*>\\s*)?[A-Za-z\\-:_]+\\s+[A-Za-z\\-:_]+"), 0) > -1)
                return tmpName;
                else break;
    }
    return name;
}

/*!
 * Implemented PerlParser
 * @see PerlParser
 */

/*!
 * @fn PerlParser::PerlParser
 * @brief Constructor
 *
 * @author Andrei Dmitriev <andrei_dm@mail.ru>
 */
PerlParser::PerlParser() {
 CppView *view = CppView::Instance();
 QListViewItem *item = view->getItem( CppView::Class );
 item->setText(0,QString("Package"));
 item->repaint();
 item = view->getItem( CppView::Macro );
 item->setText(0,QString("Constant"));
 item->repaint();
 view->FreeInst();
}

/*!
 * @fn PerlParser::parse
 * @param Kate::Document *doc
 * @return void
 * @overload AbsParser::parse
 * @author Andrei Dmitriev <andrei_dm@mail.ru>
 */
void PerlParser::parse(Kate::Document *doc)
{
    uint i = 0;
    int defDeclaration = 0;
    uint j = 0;
    int pod = 0;
    int defEnd = 0;
    enum TagType {Empty=0, Const, Func, Pack } imIN = Empty;
    QString name;
    QString curPack;
    QRegExp rxPod( "^=([A-Za-z0-9]+)" );
    QRegExp rxFunc( "^sub(\\s+|$)" );
    QRegExp _rxFunc( "^sub\\s+([A-Za-z0-9_\\-\\.]+)" );
    QRegExp _rxFuncN( "^([A-Za-z0-9_\\-\\.]+)" );
    QRegExp rxConstant( "^use\\s*constant ([A-Za-z0-9_\\-]+)" );
    QRegExp rxPackage( "^package\\s+([^;]+)" );
    for (i=0; i<doc->numLines();)
    {
            curStr = doc->textLine(i);
            curStr = curStr.replace(0x9,0x20);
            // find pod document
            if(pod) {
                       if((rxPod.search(curStr) == 0) &&
                          (rxPod.cap(1) == QString("cut")) ) pod = 0;
                       GOTO_NEXT;
            }else{
                       if((rxPod.search(curStr) == 0) &&
                          (rxPod.cap(1) != QString("cut")) ) { pod = 1; GOTO_NEXT; }

            }
            curStr = curStr.stripWhiteSpace();
            // delete all string
            if((curStr.find("\"") !=-1 || curStr.find("'")!=-1)) { // qoute
                    curStr = curStr.replace("\\\"", "");
                    curStr = curStr.replace("\\'", "");
                    curStr = curStr.replace(QRegExp("\"[^\"]*\""), "");
                    curStr = curStr.replace(QRegExp("'[^']*'"), "");
            }
            // delete comment
            int posComment = curStr.find("#");
            if(posComment != -1) curStr = curStr.left(posComment);
            switch(imIN) {
                        case Empty:
                                 defEnd = 0;
                                 name = QString("");
                                 if(rxFunc.search(curStr,0) > -1) {
                                            imIN = Func;
                                            defDeclaration = i;
                                            continue;
                                 }else if(rxConstant.search(curStr,0) > -1) {
                                            imIN = Const;
                                            continue;
                                 }else if(rxPackage.search(curStr,0) > -1) {
                                            imIN = Pack;
                                            continue;
                                 }
                        break;
                        case Func:
                                 if(_rxFunc.search(curStr) > -1) {
                                                name = _rxFunc.cap(1);
                                                if(!curPack.isEmpty()) name = curPack + name;
                                                addFunc(name, defDeclaration);
                                                imIN = Empty;
                                 }else if(defEnd){
                                                 if(_rxFuncN.search(curStr) > -1) {
                                                        name = _rxFunc.cap(1);
                                                        if(!curPack.isEmpty()) name = curPack + name;
                                                 }
                                                 addFunc(name, i);
                                                 imIN = Empty;
                                 }else defEnd = 1;
                        break;
                        case Const:
                                if(rxConstant.search(curStr) > -1) {
                                                name = rxConstant.cap(1);
                                                addMacros(name, i);
                                }
                                imIN = Empty;
                        break;
                        case Pack:
                                if(rxPackage.search(curStr) > -1) {
                                                name = rxPackage.cap(1);
                                                curPack = name + QString("::");
                                                addClass(name, i);
                                }
                                imIN = Empty;
                        break;
            }
    i++;
    }
}

/*!
 * @fn PerlParser::addFunc
 * @param QString &name
 * @param int strNum
 * @return void
 * @brief Add Function in Interface
 * @author Andrei Dmitriev <andrei_dm@mail.ru>
 */
void PerlParser::addFunc(QString &name, int strNum)
{
            QRegExp rxParse("^(.+)\\s*::\\s*([^:]+)$");
            name = name.replace("\n"," ");
            if(rxParse.search(name) != -1) {
                    QString namePack = rxParse.cap(1);
                    QString nameFunc = rxParse.cap(2);
                    ObjectMap *objMap = ObjectMap::Instance();
                    ItemObject  *objParent = objMap->findObject(namePack, CppView::Class);
                    if(!objParent) { addObject(name,strNum,CppView::Func); return; }
                    ItemObject  *obj = objMap->addObject(name, nameFunc, CppView::Func, objParent);
                    if(obj) obj->setStrNum(strNum);
                    objMap->FreeInst();
            } else addObject(name,strNum,CppView::Func);
}

void JavaScriptParser::parse(Kate::Document *doc)
{
        int i, curI, posDef, inTag, posStr, iFirst, defDeclaration = 0;
        uint j = 0;
        int comment = 0;
        int qoute = 0;
        int blockF = 0;
        int blockR = 0;
        int lastCommentStrNum = 0;
        enum TagType {Empty=0, Const, Func, Pack } imIN = Empty;
        QString name;
        QString curPack;
        QRegExp rxFunc( "^(([A-Za-z0-9_\\-\\.]+)\\s*=\\s*)?function(\\s+|\\(|$)" );
        QRegExp _rxFunc( "^function\\s+([^\\)]+\\))" );
        QRegExp _rxFuncN( "^([A-Za-z0-9_\\-\\.]+)\\s*=\\s*function\\s*(\\([^\\)]*\\))" );
        for (i=0; i<doc->numLines();)
        {
            curStr = doc->textLine(i);
            curStr = curStr.replace(0x9,0x20);
            curStr = curStr.stripWhiteSpace();
            // delete all string
            if((curStr.find("\"") !=-1 || curStr.find("'")!=-1)) { // qoute
                    curStr = curStr.replace("\\\"", "");
                    curStr = curStr.replace("\\'", "");
                    curStr = curStr.replace(QRegExp("\"[^\"]*\""), "");
                    curStr = curStr.replace(QRegExp("'[^']*'"), "");
            }
            // delete comment
        if(!comment && (curStr.find("//") != -1 || curStr.find("/*")!= -1)) { // comment
                    curStr = curStr.replace(QRegExp("/\\*.*\\*/"), "");
                    int PosOneCom = curStr.find("//");
                    posStr = curStr.find("/*");
                    if(PosOneCom != -1 && (posStr == -1 || PosOneCom < posStr )) {
                               curStr = curStr.left(PosOneCom);
                    }
                    if(posStr != -1 && ( PosOneCom == -1 || posStr < PosOneCom ) ) {
                            comment = 1;
                            curStr = curStr.left(posStr);
                    }
        } else if( comment ) {
                    if((posStr = curStr.find("*/")) != -1) {
                                shiftStr(posStr+2);
                                comment = 0;
                                lastCommentStrNum = i;
                                continue;
                    }
                    GOTO_NEXT;
        }
        switch(imIN) {
                        case Empty:
                                name = QString("");
                                 if(rxFunc.search(curStr,0) > -1) {
                                            imIN = Func;
                                            defDeclaration = i;
                                            continue;
                                 }
                        break;
                        case Func:
                                 if(_rxFunc.search(curStr) > -1) {
                                                name = _rxFunc.cap(1);
                                                addFunc(name, defDeclaration);
                                                imIN = Empty;
                                 }else if(_rxFuncN.search(curStr) > -1) {
                                                name = _rxFuncN.cap(1) + _rxFuncN.cap(2);
                                                addFunc(name, i);
                                                imIN = Empty;
                                 }else imIN = Empty;
                        break;
        }
        i++;
        }
}